from typing import Dict

from django.contrib.auth import get_user_model
from rest_framework.request import Request
from rest_framework.reverse import reverse

from bridger import display as dp
from bridger.viewsets import ModelViewSet
from bridger.serializers import ModelSerializer


def default_profile(request: Request) -> Dict:
    """Returns two endpoints, one for the profile image and one for the profile widget"""

    return {
        "image": "https://image.freepik.com/free-vector/businessman-profile-cartoon_18591-58479.jpg",
        "endpoint": reverse(
            "bridger:user-detail", args=[request.user.id], request=request
        ),
    }


class UserSerializer(ModelSerializer):
    class Meta:
        model = get_user_model()
        fields = ("id", "username", "email")


class UserViewSet(ModelViewSet):

    INSTANCE_DISPLAY = dp.InstanceDisplay(
        sections=[dp.Section(fields=dp.FieldSet(fields=["username", "email"]))]
    )

    serializer_class = UserSerializer
    queryset = get_user_model().objects.all()



# Generated by Django 3.0.5 on 2020-05-07 07:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('bridger', '0005_auto_20200402_1741'),
    ]

    operations = [
        migrations.AlterField(
            model_name='notification',
            name='endpoint',
            field=models.URLField(blank=True, max_length=2048, null=True),
        ),
    ]



# Generated by Django 3.0.6 on 2020-05-14 15:05

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('bridger', '0006_auto_20200507_0913'),
    ]

    operations = [
        migrations.CreateModel(
            name='Asset',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('file', models.FileField(upload_to='markdown/assets')),
            ],
            options={
                'verbose_name': 'Asset',
                'verbose_name_plural': 'Assets',
            },
        ),
    ]



from django.db.models import QuerySet
from rest_framework.pagination import CursorPagination as DefaultCursorPagination
from rest_framework.pagination import LimitOffsetPagination as DefaultLimitOffsetPagination
from rest_framework.pagination import BasePagination
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.views import View


class CursorPagination(DefaultCursorPagination):
    """
    A pagination class that mimics the default Django Rest Framework pagination class,
    but adds functionality to also display aggregated values
    """

    page_size = 25

    def is_initial(self):
        return self.cursor is None

    def _get_position_from_instance(self, instance, ordering):
        new_ordering = [*ordering]
        new_ordering[0] = new_ordering[0].split("__")[0]
        return super()._get_position_from_instance(instance, new_ordering)


class LimitOffsetPagination(DefaultLimitOffsetPagination):

    default_limit = 25



from bridger.viewsets import RepresentationModelViewSet
from .serializers import TagRepresentationSerializer
from .models import Tag


class TagRepresentationViewSet(RepresentationModelViewSet):
    queryset = Tag.objects.all()
    serializer_class = TagRepresentationSerializer

    search_fields = ["title"]



from rest_framework import serializers

from .mixins import BridgerSerializerFieldMixin
from .types import BridgerType


class ListField(BridgerSerializerFieldMixin, serializers.ListField):
    field_type = BridgerType.LIST.value



from typing import Dict, Iterator, List, Tuple, Union

from django.urls.exceptions import NoReverseMatch
from rest_framework.request import Request
from rest_framework.reverse import reverse

from bridger.metadata.mixins import BridgerMetadataMixin
from contextlib import suppress

Endpoint_T = Union[str, Tuple[str, List, Dict]]


class EndpointMetadata(BridgerMetadataMixin):
    key = "endpoints"
    method_name = "_get_endpoints"


class EndpointMetadataMixin:
    def get_generic_endpoint(
        self, request: Request, endpoint: Endpoint_T = None
    ) -> Endpoint_T:
        return endpoint

    def get_history_endpoint(self, request, endpoint=None):

        with suppress(AssertionError, AttributeError):
            obj = self.get_object()
            model = type(obj)
            opts = model._meta

            if hasattr(opts, "simple_history_manager_attribute") and hasattr(
                model, "get_endpoint_basename"
            ):
                return (
                    f"{model.get_endpoint_basename()}-history-list",
                    [obj.id],
                    {},
                )

        return None

    def _get_generic_endpoint(self, request: Request, endpoint_type: str) -> Endpoint_T:
        field_name = endpoint_type.upper()
        method_name = f"get_{endpoint_type}"

        endpoint = getattr(self, method_name, self.get_generic_endpoint)(
            request, getattr(self, field_name, None)
        )

        if endpoint is None:
            return endpoint

        if isinstance(endpoint, str):
            viewname, args, kwargs = endpoint, [], {}
        else:
            viewname, args, kwargs = endpoint

        try:
            return reverse(viewname=viewname, request=request, args=args, kwargs=kwargs)
        except NoReverseMatch:
            return None

    def get_pk_endpoint_field(self, request: Request) -> str:
        return "{{" + getattr(self, "PK_FIELD", "id") + "}}"

    def _get_endpoint(self, request: Request) -> Endpoint_T:
        return self._get_generic_endpoint(request, "endpoint")

    def _get_list_endpoint(self, request: Request) -> Endpoint_T:
        return self._get_generic_endpoint(request, "list_endpoint")

    def _get_instance_endpoint(self, request: Request) -> Endpoint_T:
        return self._get_generic_endpoint(request, "instance_endpoint")

    def _get_create_endpoint(self, request: Request) -> Endpoint_T:
        return self._get_generic_endpoint(request, "create_endpoint")

    def _get_delete_endpoint(self, request: Request) -> Endpoint_T:
        return self._get_generic_endpoint(request, "delete_endpoint")

    def _get_history_endpoint(self, request: Request) -> Endpoint_T:
        return self._get_generic_endpoint(request, "history_endpoint")

    def _get_endpoints(self, request: Request) -> Iterator[Dict[str, str]]:
        endpoints = dict()

        endpoint = self._get_endpoint(request=request)
        list_endpoint = self._get_list_endpoint(request=request)
        instance_endpoint = self._get_instance_endpoint(request=request)
        create_endpoint = self._get_create_endpoint(request=request)
        delete_endpoint = self._get_delete_endpoint(request=request)
        history_endpoint = self._get_history_endpoint(request=request)

        if not self.kwargs.get("pk", None):
            if url := instance_endpoint or endpoint:
                endpoints["instance"] = f"{url}{self.get_pk_endpoint_field(request)}/"

        else:
            if url := list_endpoint or endpoint:
                endpoints["list"] = url

            if history_endpoint:
                endpoints["history"] = history_endpoint

        if not getattr(self, "READ_ONLY", False):
            if url := create_endpoint or endpoint:
                endpoints["create"] = url

            if url := delete_endpoint or endpoint:
                endpoints["delete"] = f"{url}{self.get_pk_endpoint_field(request)}/"

        return endpoints



# from django.db import models

# # Create your models here.
# from bridger.signals.instance_buttons import add_instance_button
# from django.dispatch import receiver
# from tests.viewsets.model_test import ModelTestModelViewSet


# @receiver(add_instance_button, sender=ModelTestModelViewSet)
# def test_adding_instance_buttons(*args, **kwargs):
#     print("==============================")
#     print(args)
#     print(kwargs)
#     print("==============================")
#     return 1



from django.db import models
from django.db.models.fields.reverse_related import ManyToManyRel
from django.db.models.query import QuerySet


class Tag(models.Model):
    title = models.CharField(max_length=255, unique=True)
    color = models.CharField(max_length=7, default="#D3D3D3")

    @classmethod
    def get_representation_endpoint(cls):
        return "bridger:tagrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{title}}"

    def save(self, *args, **kwargs):
        self.title = self.title.lower()
        super().save(*args, **kwargs)

    def __str__(self):
        return self.title

    class Meta:
        verbose_name = "Tag"
        verbose_name_plural = "Tags"

    def get_tagged_queryset(self):
        qs = self.__class__.objects.none()
        for field in filter(lambda x: isinstance(x, ManyToManyRel), self.__class__._meta.get_fields()):
            _qs = getattr(self, field.related_name).all()
            qs = qs.union(_qs.values("tag_detail_endpoint", "tag_representation"))
        return qs


class TagModelMixin(models.Model):

    tag_detail_endpoint = models.CharField(max_length=255, null=True, blank=True)
    tag_representation = models.CharField(max_length=255, null=True, blank=True)

    tags = models.ManyToManyField(to="bridger.Tag", related_name="%(app_label)s_%(class)s_items", blank=True)

    def save(self, *args, **kwargs):
        self.tag_detail_endpoint = self.get_tag_detail_endpoint()
        self.tag_representation = self.get_tag_representation()
        super().save(*args, **kwargs)

    def get_tag_detail_endpoint(self):
        raise NotImplementedError("When using Tags, you must implement get_tag_detail_endpoint")

    def get_tag_representation(self):
        raise NotImplementedError("When using Tags, you must implement get_tag_representation")

    class Meta:
        abstract = True



from django.urls import include, path
from django.views.generic import TemplateView

from .frontend_user_configuration import FrontendUserConfigurationModelViewSet
from .notifications.viewsets import NotificationModelViewSet
from .tags.viewsets import TagRepresentationViewSet
from .routers import BridgerRouter
from .views import Config, Profile

from .share.views import ShareAPIView
from .menus.views import MenuAPIView
from .profile import UserViewSet
from .markdown.views import (
    BlockDiag,
    AssetCreateView,
    AssetRetrieveView,
    TemplateTagView,
)
from .clubhouse import ClubHouseView

router = BridgerRouter()
router.register(r"frontenduserconfiguration", FrontendUserConfigurationModelViewSet, basename="frontenduserconfiguration")
router.register(r"notification", NotificationModelViewSet, basename="notification")
router.register(r"user", UserViewSet, basename="user")
router.register(r"clubhouse", ClubHouseView, basename="clubhouse")
router.register(r"tagrepresentation", TagRepresentationViewSet, basename="tagrepresentation")

urlpatterns = [
    path("config/", Config.as_view(), name="config"),
    path("profile/", Profile.as_view(), name="profile"),
    path("menu/", MenuAPIView.as_view(), name="menu"),
    path("share/", ShareAPIView.as_view(), name="share"),
    path("markdown/blockdiag/", BlockDiag.as_view(), name="blockdiag"),
    path("markdown/asset/", AssetCreateView.as_view(), name="markdown-asset-upload"),
    path("markdown/asset/<uuid>/", AssetRetrieveView.as_view(), name="asset-retrieve"),
    path("markdown/templatetag/", TemplateTagView.as_view(), name="markdown-tags"),
    path("", include(router.urls)),
]



from typing import List

from django.contrib.contenttypes.models import ContentType
from rest_framework.request import Request
from rest_framework.reverse import reverse

from bridger.enums import Button
from bridger.metadata.mixins import BridgerMetadataMixin


class DocsMetadata(BridgerMetadataMixin):
    key = "docs"
    method_name = "_get_docs"


class DocsMetadataMixin:
    def _get_docs(self, request: Request) -> str:
        pk = self.kwargs.get("pk", None)
        endpoints = self._get_endpoints(request)

        if hasattr(self, "basename"):
            if self.basename and pk and hasattr(self, "INSTANCE_DOCS"):
                return reverse(
                    viewname=f"{self.basename}-instance-docs",
                    args=[pk],
                    request=request,
                )

            if self.basename and hasattr(self, "LIST_DOCS"):
                return reverse(viewname=f"{self.basename}-list-docs", request=request)




from rest_framework.reverse import reverse
from rest_framework import serializers

from .mixins import BridgerSerializerFieldMixin
from .types import BridgerType, ReturnContentType


class CharField(BridgerSerializerFieldMixin, serializers.CharField):
    field_type = BridgerType.TEXT.value

    def __init__(self, *args, **kwargs):
        self.secure = kwargs.pop("secure", False)
        super().__init__(*args, **kwargs)

    def get_representation(self, request, field_name):
        representation = super().get_representation(request, field_name)
        if self.secure:
            representation["secure"] = True
        return representation


class StringRelatedField(BridgerSerializerFieldMixin, serializers.StringRelatedField):
    field_type = BridgerType.TEXT.value


class TextField(CharField):
    field_type = BridgerType.TEXTEDITOR.value
    texteditor_content_type = ReturnContentType.HTML.value

    def get_representation(self, request, field_name):
        representation = super().get_representation(request, field_name)
        representation["content_type"] = self.texteditor_content_type
        return representation


class MarkdownTextField(TextField):
    field_type = BridgerType.MARKDOWNEDITOR.value
    texteditor_content_type = ReturnContentType.MARKDOWN.value

    def __init__(self, metadata_field=None, *args, **kwargs):
        self.metadata_field = metadata_field
        super().__init__(*args, **kwargs)

    def get_representation(self, request, field_name):
        representation = super().get_representation(request, field_name)
        representation["image_upload"] = reverse("bridger:markdown-asset-upload", request=request)
        representation["tags"] = reverse("bridger:markdown-tags", request=request)
        if self.metadata_field:
            representation["metadata_field"] = self.metadata_field
        return representation



from typing import List, Iterable, Dict

from rest_framework.request import Request

from bridger.buttons.bases import ButtonConfig as Button
from bridger.metadata.mixins import BridgerMetadataMixin
from bridger.utils.itertools import uniquify_dict_iterable
from bridger.signals.instance_buttons import add_instance_button


class CustomInstanceButtonMetadata(BridgerMetadataMixin):
    key = "custom_instance_buttons"
    method_name = "_get_custom_instance_buttons"


class CustomInstanceButtonMetadataMixin:
    def get_custom_list_instance_buttons(self, request: Request, buttons: List) -> List:
        return buttons

    def get_custom_instance_buttons(self, request: Request, buttons: List) -> List:
        return buttons

    def _get_custom_instance_buttons(self, request: Request) -> List:
        if "pk" in self.kwargs:
            # results = add_instance_button.send(sender=self.__class__)
            # print(results)
            buttons = self.get_custom_instance_buttons(request=request, buttons=getattr(self, "CUSTOM_INSTANCE_BUTTONS", []))
            button_list = list()
            for button in buttons:
                button.request = request
                button_list.append(dict(button))
            return uniquify_dict_iterable(button_list, "key")
        else:
            # results = add_instance_button.send(sender=self.__class__)
            # print(results)
            buttons = self.get_custom_list_instance_buttons(
                request=request, buttons=getattr(self, "CUSTOM_LIST_INSTANCE_BUTTONS", []),
            )
            button_list = list()
            for button in buttons:
                button.request = request
                button_list.append(dict(button))
            return uniquify_dict_iterable(button_list, "key")



# Generated by Django 3.0.6 on 2020-06-09 11:26

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('bridger', '0011_asset_file_url_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Tag',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=255, unique=True)),
                ('color', models.CharField(default='#D3D3D3', max_length=7)),
            ],
            options={
                'verbose_name': 'Tag',
                'verbose_name_plural': 'Tags',
            },
        ),
    ]



from rest_framework import permissions


class RestAPIModelPermissions(permissions.DjangoModelPermissions):
    """
        Mixin for adding the view permission to the perms_map
        NOTE: This is only here until Django Rest Framework patches
        their DjangoModelPermissions to include view permissions
    """

    perms_map = {
        "GET": ["%(app_label)s.view_%(model_name)s"],
        "OPTIONS": ["%(app_label)s.view_%(model_name)s"],
        "HEAD": ["%(app_label)s.view_%(model_name)s"],
        "POST": ["%(app_label)s.add_%(model_name)s"],
        "PUT": ["%(app_label)s.change_%(model_name)s"],
        "PATCH": ["%(app_label)s.change_%(model_name)s"],
        "DELETE": ["%(app_label)s.delete_%(model_name)s"],
    }



from bridger.buttons import WidgetButton
from bridger import serializers

from .models import Notification


class NotificationModelSerializer(serializers.ModelSerializer):
    @serializers.register_dynamic_button()
    def notification_buttons(self, instance, request, user):
        btns = instance.buttons or []
        if endpoint := instance.endpoint:
            if "widget_endpoint" in endpoint:
                endpoint = endpoint.split("widget_endpoint=")[1]

            btns.append(WidgetButton(endpoint=endpoint, label="Open", title="Open", icon="wb-icon-data"))
        return btns

    class Meta:
        model = Notification
        fields = read_only_fields = (
            "id",
            "recipient",
            "title",
            "message",
            "timestamp_created",
            "timestamp_received",
            "timestamp_read",
            "timestamp_mailed",
            "_buttons",
        )
